--------------------------------------------------------------------------
--  Forest library, version 1.3						--
--  Copyright (C) 1995 by Kenneth Almquist				--
--									--
--  You may redistribute or modify this software under the terms	--
--  given in the License for the Forest Library, which appears in	--
--  the file named LICENSE.						--
--------------------------------------------------------------------------

with POSIX.More,
     POSIX_IO,
     POSIX_Process_Identification;
package Forest_File_Locking is

    type Lock_Kind is (Read_Lock, Write_Lock, Unlock);

    type Lock_Length is
		range 1 .. Long_Long_Integer(POSIX_IO.IO_Offset'last) + 1;
    for Lock_Length'size use POSIX_IO.IO_Offset'size;

    function Infinity return Lock_Length;
	-- Returns lock_length'last, which stands for infinity.
    pragma inline(Infinity);


    -- A file lock is represented by file_lock record.  The start of the
    -- region is specified by Starting_Point and Start.  The length of the
    -- region is specified by Length, which may be infinite.
    type File_Lock is record
	Kind:            Lock_Kind;
	Starting_Point:  POSIX_IO.Position;
	Start:           POSIX_IO.IO_Offset;
	Length:          Lock_Length;
    end record;

    procedure Get_Lock
	    (File:     POSIX_IO.File_Descriptor;
    	     Lock:     File_Lock;
	     Result:   out File_Lock;
             Process:  out POSIX_Process_Identification.Process_ID);
	-- Exceptions:  posix_error.
	-- Sets result to "the first" lock which conflicts with lock,
	-- and process to the process_id of the process holding the
	-- lock.  If there is no conflicting lock, sets process to
	-- null_process_id.

    procedure Set_Lock
	    (File:   POSIX_IO.File_Descriptor;
	     Lock:   File_Lock);
	-- Exceptions:  posix_error.
	-- Set or release a lock, raising posix_error with error_code
	-- Resource_Temporarily_Unavailable if there is a conflicting
	-- lock.
	--
	-- If a process holds one or more locks on a file, and places
	-- a new lock on the file, the new lock overrides the old locks
	-- where the regions covered overlap.  For example, if you say
	--	set_lock(file, (write_lock, from_beginning, 3, 6);
	--	set_lock(file, (read_lock, from_beginning, 3, 2);
	-- the first set_lock call will place a write lock on the 6 byte
	-- region starting at postion 3.  The second set_lock call will
	-- place a read lock on a 2 byte region starting at position 3,
	-- and delete the write lock on this region.  The next 4 bytes
	-- will continue to be covered by the write lock from the first
	-- set_lock call.

    procedure Wait_To_Set_Lock
	    (File:	      POSIX_IO.File_Descriptor;
	     Lock:	      File_Lock;
             Masked_Signals:  POSIX.More.Signal_Handling := POSIX.RTS_Signals);
	-- Exceptions:  posix_error.
	-- Set or release a lock, blocking until any conflicting locks
	-- are released.

end Forest_File_Locking;
