/* Code to write Ada declarations.

   Copyright (C) 1995 by Kenneth Almquist

This file is part of ctoada, which is derived from GNU CC and is
distributed under the same terms and condtions.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include "config.h"
#include <stdio.h>
#include <string.h>
#include "tree.h"
#include "c-tree.h"
#include "obstack.h"


extern struct obstack permanent_obstack;

struct def {
    char *defname;
    union decl *decl_list;
    char *package;
    int indent;
    struct def *next;
};


struct c_type {
    int is_struct;
    tree name;
    tree field_name;
};


struct decl_common {
    union decl *next;
    int decl_type;
    int lineno;
};


#define ADA_TYPE 1
struct type_decl {
    union decl *next;
    int decl_type;
    int lineno;
    char *name;
    char *package;
    char *option_var;
    union decl *tnext;		/* next type */
    tree type;
    struct c_type ctype;
    char *ptrname;		/* name of pointer to this type */
    struct objdecl *field_list;
};


#define ADA_SUBR 2
struct subr_decl {
    union decl *next;
    int decl_type;
    int lineno;
    char *name;
    char *package;
    char *option_var;
    char *c_name;
    tree type;
    int is_function;
    struct objdecl *arg_list;
};


union decl {
    struct decl_common common;
    struct type_decl tdcl;
    struct subr_decl sdcl;
};


struct objdecl {
    struct objdecl *next;
    char *name;
    int lineno;
    char *type;
#ifdef TYPE_DEFINE
    char *type_define;
#endif
    tree decl;
    int unbounded;
    char *option_var;
};


#ifdef FIELD_MATCH
struct field_match {
    struct field_match *next;
    tree field;
    union decl *type;
};
#endif



char *output_file_name;		/* set in main routine */
struct def *def_list;
struct def **defpp = &def_list;
#ifdef FIELD_MATCH
struct field_match *fmatch_list;
#endif
int default_indent = 4;
char txtbuf[1024];		/* space to format using sprintf */
union decl *type_list;
union decl **end_type_list = &type_list;
FILE *outfp;
char *cur_package;
int array_sequence_number;
int record_sequence_number;


/* PACKAGE define_queue */
void queue_define();
#ifdef TYPE_DEFINE
void queue_type_define();
#endif
void put_other_defines();
/* END PACKAGE define_queue */


/* PACKAGE lexical_analysis */
#define MAXTOKLEN 1024
int gettok();
void expect();
void pushback();
void opendesc();
void closedesc();
int lasttok = -1;
char *toktext;
extern int lineno;
char *types_file_name;	/* set by main routine */

/* Token types */
#define TEOF 1		/* end of file */
#define TNL 2		/* newline */
#define TNUMBER 3	/* number */
#define TIDENT 4	/* identifier */
#define TLP 5		/* left paren '(' */
#define TRP 6		/* right paren ')' */
#define TDOT 7		/* period '.' */
#define TCOMMA 8	/* comma ',' */
/* END PACKAGE lexical_analysis */


/* PACKAGE output */
void start_output();
void push_output();
void pop_output();
void newline_and_indent();
void outs();
void outc();
/* END PACKAGE output */


#define equal(s1, s2)	(strcmp(s1, s2) == 0)

void parse();
void parse_define();
void parse_types();
union decl *get_decl();
void parse_fields();
struct objdecl *parse_field();
union decl *get_subr_decl();
void c_type_name();
tree lookup_c_type();

void put_types();
void put_type();
void put_subr();
void put_field_name();
int is_basic_type();
char *basic_type();
char *savestr();
void err();

long strtol();



void
write_ada_definitions() {
    parse();
    put_types();
}



void
parse() {
    int t;

    opendesc();
    while ((t = gettok()) != TEOF) {
	if (t == TIDENT && equal(toktext, "define")) {
	    parse_define();
	} else if (t == TNL) {
	    /* ignore blank line */
	} else {
	    err("Expecting \"define\"");
	    while (t != TNL && t != TEOF)
		t = gettok();
	    pushback();
	}
    }
    closedesc();
}


void
parse_define() {
    struct def *defp;
    int t;

    defp = (struct def *) xmalloc(sizeof (struct def));
    expect(TIDENT, "expecting name to define");
    defp->defname = savestr(toktext);
    defp->package = NULL;
    defp->indent = default_indent;
    if ((t = gettok()) != TNL) {
	for (;;) {
	    if (equal(toktext, "package")) {
		for (;;) {
		    expect(TIDENT, "expecting package name");
		    obstack_grow(&permanent_obstack, toktext, strlen(toktext));
		    if (gettok() != TDOT)
			break;
		    obstack_1grow(&permanent_obstack, '.');
		}
		pushback();
		obstack_1grow(&permanent_obstack, '\0');
		defp->package = obstack_finish(&permanent_obstack);
	    } else if (equal(toktext, "indent")) {
		expect(TNUMBER, "expecting indentation value");
		defp->indent = strtol(toktext, (char **) NULL, 0);
	    } else {
		err("Unexpected token");
	    }
	    if ((t = gettok()) == TNL)
		break;
	    if (t != TCOMMA)
		err("Expecting comma or newline");
	    if (gettok() == TNL)
		gettok();
	}
    }
    defp->next = NULL;
    *defpp = defp;
    defpp = &defp->next;
    parse_types(defp);
}



void
parse_types(defp)
    struct def *defp;
{
    int t;
    union decl **dpp = &defp->decl_list;
    union decl *declp;

    defp->decl_list = NULL;
    for (;;) {
	declp = NULL;
	if (gettok() == TNL) {
	    /* ignore blank line */
	} else if (equal(toktext, "type")) {
	    declp = get_decl(defp, 0);
	} else if (equal(toktext, "record")) {
	    declp = get_decl(defp, 1);
	    parse_fields(declp);
	} else if (equal(toktext, "procedure")) {
	    declp = get_subr_decl(defp, 0);
	} else if (equal(toktext, "function")) {
	    declp = get_subr_decl(defp, 1);
	} else if (equal(toktext, "end")) {
	    if (gettok() != TIDENT || ! equal(toktext, "define"))
		err("Expecting \"end define\"");
	    else if (gettok() != TNL)
		err("Expecting newline");
	    break;
	} else {
	    err("expecting keyword (\"type\", \"record\", etc.)");
	}
	if (declp) {
	    declp->common.next = NULL;
	    *dpp = declp;
	    dpp = &declp->common.next;
	    if (declp->common.decl_type == ADA_TYPE) {
		*end_type_list = declp;
		end_type_list = &declp->tdcl.tnext;
		declp->tdcl.tnext = NULL;
	    }
	}
    }
}


/*
 * Read the first line of a type or record declaration.
 */
union decl *
get_decl(defp, is_record)
    struct def *defp;
    int is_record;
{
    int optional;
    int t;
    tree type;
    union decl *declp = (union decl *) xmalloc(sizeof (struct type_decl));

    declp->common.decl_type = ADA_TYPE;
    declp->common.lineno = lineno;
    expect(TIDENT, "expecting type name");
    declp->tdcl.name = savestr(toktext);
    declp->tdcl.package = defp->package;
    declp->tdcl.field_list = NULL;
    declp->tdcl.ptrname = NULL;
    optional = 0;
    declp->tdcl.ctype.name = NULL_TREE;
    declp->tdcl.ctype.field_name = NULL_TREE;
    declp->tdcl.option_var = NULL;
    if (gettok() != TNL) {
	for (;;) {
	    if (equal(toktext, "is")) {
		c_type_name(&declp->tdcl.ctype);
		declp->tdcl.ctype.field_name = NULL_TREE;
	    } else if (equal(toktext, "matches")) {
		expect(TIDENT, "expecting field name");
		declp->tdcl.ctype.field_name = get_identifier(toktext);
		if (gettok() != TIDENT || ! equal(toktext, "in"))
		    err("Expecting \"in\"");
		c_type_name(&declp->tdcl.ctype);
	    } else if (equal(toktext, "optional")) {
		optional = 1;
		if (gettok() == TIDENT)
		    declp->tdcl.option_var = savestr(toktext);
		else
		    pushback();
	    } else if (equal(toktext, "access_type")) {
		expect(TIDENT, "expecting access type name");
		declp->tdcl.ptrname = savestr(toktext);
	    } else {
		err("Unimplemented option");
	    }
	    if ((t = gettok()) == TNL)
		break;
	    if (t != TCOMMA)
		err("Expecting comma or newline");
	    if (gettok() == TNL)
		gettok();
	}
    }
    if (declp->tdcl.ctype.name == NULL_TREE) {
	declp->tdcl.ctype.name = get_identifier(declp->tdcl.name);
	declp->tdcl.ctype.is_struct = 0;
    }
    type = lookup_c_type(&declp->tdcl.ctype, declp, ! optional);
    if (type != NULL && is_record) {
	if (TREE_CODE(type) == TYPE_DECL)
	    type = TREE_TYPE(type);
	if (TREE_CODE(type) != RECORD_TYPE) {
	    err("Not a record type");
	    type = NULL;
	}
    }
    declp->tdcl.type = type;
    return declp;
}


void
parse_fields(declp)
    union decl *declp;
{
    struct objdecl **fldpp = &declp->tdcl.field_list;
    struct objdecl *fldp;

    for (;;) {
	if (gettok() == TNL)
	    continue;
	if (equal(toktext, "field")) {
	    fldp = parse_field(declp);
	    *fldpp = fldp;
	    fldpp = &fldp->next;
	} else if (equal(toktext, "end")) {
	    if (gettok() != TIDENT || ! equal(toktext, "record"))
		err("Expecting \"end record\"");
	    else if (gettok() != TNL)
		err("Expecting newline");
	    break;
	} else {
	    err("expecting \"field\"");
	}
    }
    *fldpp = NULL;
#if 0    /* DEBUG */
    fprintf(stderr, "field list:");
    for (fldp = declp->tdcl.field_list ; fldp ; fldp = fldp->next)
	fprintf(stderr, " %s", fldp->name);
    putc('\n', stderr);
#endif
}


struct objdecl *
parse_field(declp)
    union decl *declp;
{
    struct objdecl *fldp;
    tree fptr;
    char *fldname;
    int optional;
    int t;

    expect(TIDENT, "expecting field name");
    fldp = (struct objdecl *) xmalloc(sizeof (struct objdecl));
    fldp->name = savestr(toktext);
    fldp->lineno = lineno;
    fldp->type = NULL;
#ifdef TYPE_DEFINE
    fldp->type_define = NULL;
#endif
    fldp->option_var = NULL;
    fldp->unbounded = 0;
    if (gettok() != TNL) {
	for (;;) {
	    if (equal(toktext, "optional")) {
		optional = 1;
		if (gettok() == TIDENT)
		    fldp->option_var = savestr(toktext);
		else
		    pushback();
	    } else if (equal(toktext, "unbounded")) {
		fldp->unbounded = 1;
	    } else if (equal(toktext, "type")) {
		expect(TIDENT, "expecting type name");
		fldp->type = savestr(toktext);
#ifdef TYPE_DEFINE
	    } else if (equal(toktext, "type_define")) {
		expect(TIDENT, "expecting name");
		fldp->type_define = savestr(toktext);
#endif
	    } else {
		err("Unimplemented option");
	    }
	    if ((t = gettok()) == TNL)
		break;
	    if (t != TCOMMA)
		err("Expecting comma or newline");
	    if (gettok() == TNL)
		gettok();
	}
    }
    if (declp->tdcl.type) {
	for (fptr = TYPE_FIELDS(declp->tdcl.type) ; ; fptr = TREE_CHAIN(fptr)) {
	    if (fptr == NULL_TREE) {
		break;
	    }
	    fldname = IDENTIFIER_POINTER(DECL_NAME(fptr));
	    if (equal(fldname, fldp->name))
		break;
	}
    } else {
	fptr = NULL;
    }
    fldp->decl = fptr;
    if (fptr == NULL && ! optional) {
	sprintf(txtbuf, "No such field: %s", fldp->name);
	err(txtbuf);
    }
    return fldp;
}


union decl *
get_subr_decl(defp, is_function)
    struct def *defp;
    int is_function;
{
    int t;
    struct objdecl *argp;
    struct objdecl **argpp;
    char *name;
    tree cdecl;
    union decl *declp = (union decl *) xmalloc(sizeof (struct subr_decl));
    declp->common.decl_type = ADA_SUBR;
    declp->common.lineno = lineno;
    expect(TIDENT, "expecting type name");
    declp->sdcl.name = name = savestr(toktext);
    declp->sdcl.package = defp->package;
    declp->sdcl.is_function = is_function;
    if (gettok() == TLP) {
	argpp = &declp->sdcl.arg_list;
	if ((t = gettok()) != TRP) {
	    for (;;) {
		if (t != TIDENT)
		    err("Expecting argument name");
		argp = (struct objdecl *) xmalloc(sizeof (struct objdecl));
		argp->lineno = lineno;
		argp->name = savestr(toktext);
		argp->decl = NULL_TREE;
		argp->unbounded = 0;
		*argpp = argp;
		argpp = &argp->next;
		if ((t = gettok()) == TRP)
		    break;
		else if (t != TCOMMA)
		    err("Expecting comma");
	    }
	}
	*argpp = NULL;
    } else {
	pushback();
    }
    if (gettok() != TNL)
	err("Expecting newline");
    declp->sdcl.c_name = name;
    cdecl = lookup_name(get_identifier(name));
    if (cdecl == NULL_TREE) {
	sprintf(txtbuf, "%s is not defined", name);
	err(txtbuf);
    } else if (TREE_CODE(cdecl) != FUNCTION_DECL) {
	fprintf(stderr, "cdecl tree:");
	debug_tree(cdecl);
	sprintf(txtbuf, "%s is not a C function", name);
	err(txtbuf);
    }
    declp->sdcl.type = cdecl;
    return declp;
}



/*
 * Read the name of a C type and return a pointer to the definition.
 */
void
c_type_name(tp)
    struct c_type *tp;
{
    if (gettok() != TIDENT)
	err("Expecting type name");
    if (equal(toktext, "struct")) {
	expect(TIDENT, "expecting structure tag");
	tp->is_struct = 1;
    } else {
	tp->is_struct = 0;
    }
    tp->name = get_identifier(toktext);
}


/*
 * Look up a C type.  If error_flag is true, print an error message if the
 * type is not found.
 */
tree
lookup_c_type(tp, declp, error_flag)
    struct c_type *tp;
    union decl *declp;
    int error_flag;
{
    tree decl;
    tree fptr;
#ifdef FIELD_MATCH
    struct field_match *fmatch;
#endif

    if (! tp->is_struct) {
	decl = lookup_name(tp->name);
    } else {
	decl = xref_tag(RECORD_TYPE, tp->name);
	if (TREE_CODE(decl) != RECORD_TYPE || TYPE_SIZE(decl) == NULL_TREE)
	    decl = NULL_TREE;
    }
    if (decl == NULL_TREE && error_flag) {
	sprintf(txtbuf, "%s %s is not defined",
		tp->is_struct? "struct" : "type",
		IDENTIFIER_POINTER(tp->name));
	err(txtbuf);
    }
    if (tp->field_name != NULL_TREE) {
	fptr = NULL_TREE;
	if (TREE_CODE(decl) == RECORD_TYPE) {
	    for (fptr = TYPE_FIELDS(decl) ;
		 fptr != NULL_TREE && DECL_NAME(fptr) != tp->field_name ;
		 fptr = TREE_CHAIN(fptr));
	}
	if (fptr == NULL && error_flag) {
	    sprintf(txtbuf, "%s %s does not contain field %s",
		    tp->is_struct? "struct" : "type",
		    IDENTIFIER_POINTER(tp->name),
		    IDENTIFIER_POINTER(tp->field_name));
	    err(txtbuf);
	}
	decl = TREE_TYPE(fptr);
#ifdef FIELD_MATCH
	fmatch = (struct field_match *) xmalloc(sizeof (struct field_match));
	fmatch->field = fptr;
	fmatch->type = declp;
	fmatch->next = fmatch_list;
	fmatch_list = fmatch;
#endif
    }
    return decl;
}



tree
lookup_c_struct(name)
    char *name;
{
    tree ident;
    tree decl;

    ident = get_identifier(name);
    decl = xref_tag(RECORD_TYPE, ident);
    if (TREE_CODE(decl) != RECORD_TYPE || TYPE_SIZE(decl) == NULL_TREE) {
	sprintf(txtbuf, "struct %s is not defined", name);
	err(txtbuf);
    }
    return decl;
}



int indent;


void
put_types() {
    struct def *defp;
    union decl *declp;

    if ((outfp = fopen(output_file_name, "w")) == NULL) {
	fprintf(stderr, "Cannot create %s\n", output_file_name);
	exit(2);
    }
    for (defp = def_list ; defp ; defp = defp->next) {
	cur_package = defp->package;
	fprintf(outfp, "define([%s], [", defp->defname);
	start_output(defp->indent);
	for (declp = defp->decl_list ; declp ; declp = declp->common.next) {
	    lineno = declp->common.lineno;
	    push_output();
	    if (declp->common.decl_type == ADA_TYPE) {
		if (declp->tdcl.type != NULL_TREE) {
		    sprintf(txtbuf, "type %s is ", declp->tdcl.name);
		    outs(txtbuf);
		    put_type(declp->tdcl.type, declp, 0);
		    outc(';');
		    if (declp->tdcl.ptrname) {
			newline_and_indent();
			sprintf(txtbuf, "type %s is access all %s;",
				    declp->tdcl.ptrname, declp->tdcl.name);
			outs(txtbuf);
		    }
		}
		if (declp->tdcl.option_var) {
		    queue_define(declp->tdcl.option_var,
				 declp->tdcl.type? "1" : "0");
		}
	    } else {
		put_subr(declp);
	    }
	    pop_output();
	}
	fputs("])\n", outfp);
	put_other_defines();
    }
}



char *ic_string_type = "POSIX.POSIX_String";
char *ic_address = "System.Address";

/*
 * Write out a type.  The argument can be either a type or a typedef.
 * The second argument is non-null if we are defining an Ada type.
 */
void
put_type(type, declp, unbounded)
    tree type;
    union decl *declp;
    int unbounded;
{
    union decl *dp;
#ifdef FIELD_MATCH
    struct field_match *fmatch;
#endif
    tree tname;
    tree minval;
    tree fptr;
    tree ftype;
    char *fldname;
    char *p;
    tree array_last;
    struct objdecl *finfo;
    static struct objdecl dummy_field;
    int need_pad;
    int bitpos;
    tree bitpos_tree;
    tree field_size;
    int pad_sequence_number;

    if (TREE_CODE(type) == TYPE_DECL) {
	type = TREE_TYPE(type);
    }

    if (declp == NULL && ! is_basic_type(type)) {
	tree ptr_to, named_type;

	ptr_to = NULL_TREE;
	if (TREE_CODE(type) == POINTER_TYPE)
	    ptr_to = TREE_TYPE(type);
	for (dp = type_list ; dp ; dp = dp->tdcl.tnext) {
	    named_type = dp->tdcl.type;
	    if (named_type == NULL_TREE)
		continue;
	    if (TREE_CODE(named_type) == TYPE_DECL) {
		named_type = TREE_TYPE(named_type);
	    }
	    if (named_type == type) {
		ptr_to = NULL_TREE;
		break;
	    }
	    if (named_type == ptr_to && dp->tdcl.ptrname != NULL) {
		if (ptr_to == NULL_TREE)
		    abort();
		break;
	    }
	}
	/* We have found a match.  If ptr_to is not null, we have matched
	   an access type */
	if (dp) {
	    if (dp->tdcl.package
	     && (cur_package == NULL || ! equal(dp->tdcl.package, cur_package))) {
		outs(dp->tdcl.package);
		outc('.');
	    }
	    outs(ptr_to != NULL_TREE? dp->tdcl.ptrname : dp->tdcl.name);
	    return;
	}
    }

    if (TREE_CODE(type) == IDENTIFIER_NODE) {
	if (declp)
	    outs("new ");
	outs(IDENTIFIER_POINTER(type));
	outs(" <-- should not happen");
    } else if (TREE_CODE(type) == TYPE_DECL) {
	if (declp)
	    outs("new ");
	outs(IDENTIFIER_POINTER(DECL_NAME(type)));
	outs(" <-- should not happen");
    } else if (TREE_CODE(type) == INTEGER_TYPE) {
	if (declp)
	    outs("new ");
	minval = TYPE_MIN_VALUE(type);
	tname = TREE_TYPE(minval);
	if ((p = basic_type(tname)) != NULL) {
	    outs(p);
	} else {
	    outs("Unknown integer type");
	    debug_tree(tname);
	}
    } else if (TREE_CODE(type) == RECORD_TYPE) {
	char type_name_buf[20];
	char *type_name;
	if (declp == NULL) {
	    sprintf(type_name_buf, "record_type_%d", ++record_sequence_number);
	    type_name = type_name_buf;
	    outs(type_name);
	    push_output();
	    sprintf(txtbuf, "type %s is ", type_name);
	    outs(txtbuf);
	} else {
	    type_name = declp->tdcl.name;
	}
	outs("record");
	indent += 4;
	bitpos = 0;
	need_pad = 0;
	pad_sequence_number = 0;
	for (fptr = TYPE_FIELDS(type) ; fptr != NULL_TREE ;
					fptr = TREE_CHAIN(fptr)) {
	    if (TREE_CODE(fptr) != FIELD_DECL)
		abort();
	    for (finfo = declp? declp->tdcl.field_list : NULL ;
		 finfo != NULL && finfo->decl != fptr ; finfo = finfo->next);
	    if (finfo == NULL) {
		finfo = &dummy_field;
		if (DECL_NAME(fptr) == NULL_TREE)
		    dummy_field.name = "";	/* unnamed bit field */
		else
		    dummy_field.name = IDENTIFIER_POINTER(DECL_NAME(fptr));
	    }
	    if (DECL_BIT_FIELD_TYPE(fptr)) {	/* if a bit field */
		need_pad = 1;
	    } else {
		if (need_pad) {
		    bitpos_tree = DECL_FIELD_BITPOS(fptr);
		    if (bitpos >= 0 && TREE_CODE(bitpos_tree) == INTEGER_CST) {
			newline_and_indent();
			sprintf(txtbuf, "bit_field_substitute_%d: string (1 .. %d);",
				++pad_sequence_number,
				TREE_INT_CST_LOW(bitpos_tree) / BITS_PER_UNIT
				 - bitpos / BITS_PER_UNIT);
			outs(txtbuf);
		    }
		}
		dp = NULL;
		if (finfo->type) {
		    for (dp = type_list ; ; dp = dp->tdcl.tnext) {
			if (dp == NULL) {
			    fprintf(stderr, "%s:%d: Undefined Ada type %s\n",
				    types_file_name, finfo->lineno, finfo->type);
			    exit(2);
			}
			if (equal(dp->tdcl.name, finfo->type))
			    break;
		    }
		    if (! same_type(TREE_TYPE(fptr), dp->tdcl.type)) {
			fprintf(stderr, "%s:%d: Actual field type does not match %s\n",
				types_file_name, finfo->lineno, finfo->type);
			exit(2);
		    }
		}
#ifdef FIELD_MATCH
		fmatch = NULL;
		if (dp == NULL) {
		    for (fmatch = fmatch_list ; fmatch ; fmatch = fmatch->next) {
			if (fmatch->field == fptr) {
			    fprintf(stderr, "Field matched:  type=%s\n",
					fmatch->type->tdcl.name);
			    break;
			}
		    }
		}
#endif
		ftype = dp? dp->tdcl.type : TREE_TYPE(fptr);
		newline_and_indent();
		put_field_name(finfo->name, 12);
#ifdef FIELD_MATCH
		if (fmatch)
		    outs(fmatch->type->tdcl.name);
		else
		    put_type(ftype, NULL, finfo->unbounded);
#else
		if (dp  == NULL)
		    put_type(ftype, NULL, finfo->unbounded);
		else {
		    if (dp->tdcl.package
		     && (cur_package == NULL
			 || ! equal(dp->tdcl.package, cur_package))) {
			outs(dp->tdcl.package);
			outc('.');
		    }
		    outs(dp->tdcl.name);
		}
#endif
		outc(';');
#ifdef TYPE_DEFINE
		if (finfo->type_define)
		    queue_type_define(finfo->type_define, ftype, finfo->unbounded);
#endif
		bitpos_tree = DECL_FIELD_BITPOS(fptr);
		field_size = DECL_SIZE(fptr);
		if (TREE_CODE(bitpos_tree) == INTEGER_CST
		 && TREE_CODE(field_size) == INTEGER_CST) {
		    bitpos = TREE_INT_CST_LOW(bitpos_tree) 
			   + TREE_INT_CST_LOW(field_size);
		} else {
		    bitpos = -1;
		}
		need_pad = 0;
	    }
	}
	if (need_pad) {
	    bitpos_tree = TYPE_SIZE(type);
	    if (bitpos >= 0 && TREE_CODE(bitpos_tree) == INTEGER_CST) {
		newline_and_indent();
		sprintf(txtbuf, "bit_field_substitute_%d: string (1 .. %d);",
			++pad_sequence_number,
			TREE_INT_CST_LOW(bitpos_tree) / BITS_PER_UNIT
			 - bitpos / BITS_PER_UNIT);
		outs(txtbuf);
	    }
	}
	indent -= 4;
	newline_and_indent();
	outs("end record;");
	newline_and_indent();
	sprintf(txtbuf, "pragma convention(C, %s)", type_name);
	outs(txtbuf);
	if (declp == NULL) {
	    outc(';');
	    pop_output();
	}
	for (finfo = declp? declp->tdcl.field_list : NULL ;
	     finfo != NULL ; finfo = finfo->next) {
	    if (finfo->option_var) {
		queue_define(finfo->option_var, finfo->decl? "1" : "0");
	    }
	}
    } else if (TREE_CODE(type) == ARRAY_TYPE) {
	array_last = TYPE_MAX_VALUE(TYPE_DOMAIN(type));
	if (TREE_TYPE(type) == char_type_node) {
	    if (unbounded) {
		sprintf(txtbuf, "%s (1 .. Integer'last)", ic_string_type);
	    } else {
		sprintf(txtbuf, "%s (1 .. %ld)", ic_string_type,
			(long) TREE_INT_CST_LOW(array_last) + 1);
	    }
	    outs(txtbuf);
	} else if (declp) {
	    sprintf(txtbuf, "array (integer range 0 .. %ld) of ",
			(long) TREE_INT_CST_LOW(array_last));
	    outs(txtbuf);
	    put_type(TREE_TYPE(type), NULL, 0);
	} else {
	    char type_name[16];
	    push_output();
	    sprintf(type_name, "array_type_%d", ++array_sequence_number);
	    sprintf(txtbuf, "type %s is array (integer range 0 .. %ld) of ",
			type_name, (long) TREE_INT_CST_LOW(array_last));
	    outs(txtbuf);
	    put_type(TREE_TYPE(type), NULL, 0);
	    outc(';');
	    pop_output();
	    outs(type_name);
	}
    } else if (TREE_CODE(type) == POINTER_TYPE) {
	if (declp) {
	    outs("access ");
	    put_type(TREE_TYPE(type), NULL, 0);
	} else {
	    outs(ic_address);
	}
    } else {
	outs("Bad tree");
	debug_tree(type);
    }
}


void
put_subr(declp)
    union decl *declp;
{
    tree fdecl = declp->sdcl.type;
    tree ftype = TREE_TYPE(fdecl);
    tree parm;
    tree parmtype;
    char argname[20];
    int argnum;

    sprintf(txtbuf, "%s %s",
	    declp->sdcl.is_function? "function" : "procedure",
	    declp->sdcl.name);
    outs(txtbuf);
    parm = TYPE_ARG_TYPES(ftype);
    argnum = 0;
    if (parm != NULL_TREE) {
	outc('(');
	for (;;) {
	    sprintf(argname, "arg%d", ++argnum);
	    put_field_name(argname, 1);
	    if (TREE_CODE(parm) != TREE_LIST)
		abort();
	    parmtype = TREE_VALUE(parm);
	    put_type(parmtype, NULL, 0);
	    parm = TREE_CHAIN(parm);
	    if (parm == NULL_TREE)
		abort();
	    parmtype = TREE_VALUE(parm);
	    if (parmtype == void_type_node)
		break;
	    outs("; ");
	}
	outc(')');
    }
    if (declp->sdcl.is_function) {
	fprintf(stderr, "Result for function %s:\n", declp->sdcl.name);
	debug_tree(TREE_TYPE(ftype));
	outs(" return ");
	put_type(TREE_TYPE(ftype), NULL, 0);
    }
    outc(';');
    newline_and_indent();
    sprintf(txtbuf, "pragma import(C, %s, \"%s\");",
		declp->sdcl.name, declp->sdcl.c_name);
    outs(txtbuf);
}


void
put_field_name(name, minwidth)
    char *name;
    int minwidth;
{
    register char *p;
    int count;

    p = name;
    count = 0;
    if (*p == '_') {
	outc('X');
	count++;
    }
    while (*p != '\0') {
	outc(*p);
	count++;
	if (*p == '_' && (p[1] == '_' || p[1] == '\0')) {
	    outc('X');
	    count++;
	}
	p++;
    }
    outc(':');
    do {
	outc(' ');
    } while (++count < minwidth);
}


int
is_basic_type(type)
    tree type;
{
    return type == integer_type_node
	|| type == short_integer_type_node
	|| type == long_integer_type_node
	|| type == unsigned_type_node
	|| type == short_unsigned_type_node
	|| type == long_unsigned_type_node
	|| type == char_type_node
	|| type == signed_char_type_node
	|| type == unsigned_char_type_node;
}


char *ic_integer_type = "Interfaces.C.int";
char *ic_short_integer_type = "Interfaces.C.short";
char *ic_long_integer_type = "Interfaces.C.long";
char *ic_unsigned_type = "Interfaces.C.unsigned";
char *ic_short_unsigned_type = "Interfaces.C.unsigned_short";
char *ic_long_unsigned_type = "Interfaces.C.unsigned_long";
char *ic_char_type = "Interfaces.C.plain_char";
char *ic_signed_char_type = "Interfaces.C.signed_char";
char *ic_unsigned_char_type = "Interfaces.C.unsigned_char";


/*
 * Return a string describing a basic type.
 */
char *
basic_type(type)
    tree type;
{
    if (type == integer_type_node)
	return ic_integer_type;
    else if (type == short_integer_type_node)
	return ic_short_integer_type;
    else if (type == long_integer_type_node)
	return ic_long_integer_type;
    else if (type == unsigned_type_node)
	return ic_unsigned_type;
    else if (type == short_unsigned_type_node)
	return ic_short_unsigned_type;
    else if (type == long_unsigned_type_node)
	return ic_long_unsigned_type;
    else if (type == char_type_node)
	return ic_char_type;
    else if (type == signed_char_type_node)
	return ic_signed_char_type;
    else if (type == unsigned_char_type_node)
	return ic_unsigned_char_type;
    else
	return NULL;
}


/*
 * Return true if both types are structurally the same.  Incomplete.
 */
int
same_type(t1, t2)
    tree t1, t2;
{
    if (TREE_CODE(t1) == TYPE_DECL)
	t1 = TREE_TYPE(t1);
    if (TREE_CODE(t2) == TYPE_DECL)
	t2 = TREE_TYPE(t2);
    if (t1 == t2)
	return 1;
#if 0
    fprintf(stderr, "*** same_type:\n");
    debug_tree(t1);
    debug_tree(t2);
#endif
    if (TREE_CODE(t1) != TREE_CODE(t2))
	return 0;
    if (TREE_CODE(t1) == INTEGER_TYPE) {
	return TYPE_MIN_VALUE(t1) == TYPE_MIN_VALUE(t2);
    } else {
	return 0;
    }
}



/*
 * Return a copy of a string in permanent storage.
 */
char *
savestr(s)
    char *s;
{
    int len = strlen(s);
    return obstack_copy0(&permanent_obstack, s, len);
}



/* PACKAGE BODY lexical_analysis */

#define NO_PEEK -2

char token_text[2][MAXTOKLEN];
int pushed_token = -1;
int prev_tok;
FILE *descfp;
int peekc = NO_PEEK;
void too_long();


/*
 * Open the description file.
 */
void
opendesc() {
    if ((descfp = fopen(types_file_name, "r")) == NULL) {
	fprintf(stderr, "Cannot open %s\n", types_file_name);
	exit(2);
    }
    lineno = 1;
}


/*
 * Close the description file.
 */
void
closedesc() {
    fclose(descfp);
}


/*
 * Read the next token, and complain if it is not what is expected.
 */
void
expect(token, msg)
    int token;
    char *msg;
{
    if (gettok() != token)
	err(msg);
}


/*
 * Read the next lexical token from the description file.  The type of
 * token is returned and the text of the token is pointed to by toktext.
 * The return value is stored in the global variable lasttok for convenience.
 */
int
gettok() {
    register int c;
    register char *p;
    char *endtok;
    int tok;

    if (toktext == token_text[0])
	toktext = token_text[1];
    else
	toktext = token_text[0];
    prev_tok = lasttok;
    if (pushed_token >= 0) {
	lasttok = pushed_token;
	pushed_token = -1;
	return lasttok;
    }
    if (lasttok == TNL)
	lineno++;
    endtok = toktext + MAXTOKLEN - 1;
    if ((c = peekc) == NO_PEEK)
	c = getc(descfp);
    peekc = NO_PEEK;
    while (c == ' ' || c == '\t')
	c = getc(descfp);
do_switch:
    p = toktext;
    *p++ = c;
    switch (c) {
	case '-':
	    if ((c = getc(descfp)) == '-') {
		/* got a comment */
		while ((c = getc(descfp)) != '\n' && c != EOF);
		goto do_switch;
	    } else {
		goto badchar;	/* character not recognized */
	    }
	    break;
	case '0':
	    /* octal or hex number */
	    c = getc(descfp);
	    if (c == 'X' || c == 'x') {
		*p++ = c;
		for (;;) {
		    c = getc(descfp);
		    if ((c < '0' || c > '9')
		     && (c < 'a' || c > 'f')
		     && (c < 'A' || c > 'F'))
			break;
		    if (p >= endtok) {
			err("Token too long");
			break;
		    }
		    *p++ = c;
		}
		if (p == toktext + 2)
		    err("Bad hex number: 0x");
		else if (c >= 'g' && c <= 'z' || c >= 'G' && c <= 'Z')
		    err("Bad hex digit");
	    } else {
		while (c >= '0' && c <= '7') {
		    if (p >= endtok) {
			err("Token too long");
			break;
		    }
		    *p++ = c;
		    c = getc(descfp);
		}
	    }
	    peekc = c;
	    *p = '\0';
	    tok = TNUMBER;
	    break;
	case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	    while ((c = getc(descfp)) >= '0' && c <= '9') {
		if (p >= endtok) {
		    err("Token too long");
		    break;
		}
		*p++ = c;
	    }
	    peekc = c;
	    *p = '\0';
	    tok = TNUMBER;
	    break;
	case '(':
	    tok = TLP;
	    break;
	case ')':
	    tok = TRP;
	    break;
	case '.':
	    tok = TDOT;
	    break;
	case ',':
	    tok = TCOMMA;
	    break;
	case '\n':
	    tok = TNL;
	    break;
	case EOF:
	    p = toktext;	/* text will be the empty string */
	    tok = TEOF;
	    break;
	default:
	    /* check for identifier */
	    if (c >= 'a' && c <= 'z'
	     || c >= 'A' && c <= 'Z'
	     || c == '_' || c == '$') {
		while ((c = getc(descfp)) >= 'a' && c <= 'z'
				     || c >= 'A' && c <= 'Z'
				     || c == '_' || c == '$'
				     || c >= '0' && c <= '9') {
		    if (p >= endtok) {
			err("Token too long");
			break;
		    }
		    *p++ = c;
		}
		peekc = c;
		tok = TIDENT;
	    } else {
badchar:
		sprintf(txtbuf, "Illegal character (octal %o)", c);
		*p = '\0';
		err(txtbuf);
		goto do_switch;
	    }
	    break;
    }
    *p = '\0';
    return lasttok = tok;
}


/*
 * Undo the last call to gettok.  Only one token can be pushed back.
 */
void
pushback() {
    if (pushed_token != -1)
	abort();	/* pushback already called */
    if (lasttok == -1)
	abort();	/* no token to push back */
    if (toktext == token_text[0])
	toktext = token_text[1];
    else
	toktext = token_text[0];
    pushed_token = lasttok;
    lasttok = prev_tok;
}

void
err(msg)
    char *msg;
{
    char *t;
    fprintf(stderr, "%s:%d: %s\n", types_file_name, lineno, msg);
    if (lasttok == TNL)
	t = "NEWLINE";
    else if (lasttok == TEOF)
	t = "EOF";
    else
	t = toktext;
    fprintf(stderr, "Input token is %s\n", t);
    exit(2);
}

/* END PACKAGE lexical_analysis */


/* PACKAGE BODY output */

/*
 * This package keeps a stack of outputs.  This is useful if we are
 * in the middle of outputting a declaration and discover we need to
 * output a type declaration.
 */

struct outdata {
    struct outdata *next;
    int len;
    char buf[20];	/* increase this */
};

struct stackout {
    struct stackout *prev;
    struct outdata *data;
    int save_indent;
};

int base_indent;
struct stackout *osp;
int at_start;


void
start_output(ind)
    int ind;
{
    base_indent = indent = ind;
    at_start = 1;
}

void
push_output() {
    struct stackout *p;

    p = (struct stackout *) xmalloc(sizeof (struct stackout));
    p->prev = osp;
    osp = p;
    p->data = NULL;
    p->save_indent = indent;
    indent = base_indent;
}

void
pop_output() {
    struct stackout *sp = osp;
    struct outdata *odp, *next, *list;
    int col;

    if (at_start) {
	at_start = 0;
    } else {
	putc('\n', outfp);
	for (col = 0 ; col < base_indent ; col++)
	    putc(' ', outfp);
    }
    list = NULL;
    for (odp = sp->data ; odp ; odp = next) {
	next = odp->next;
	odp->next = list;
	list = odp;
    }
    for (odp = list ; odp ; odp = next) {
	next = odp->next;
	fwrite(odp->buf, odp->len, 1, outfp);
	free((char *) odp);
    }
    indent = sp->save_indent;
    osp = sp->prev;
    free((char *) sp);
}


void
newline_and_indent() {
    int i;

    if (osp->data != NULL) {
	outc('\n');
	for (i = indent ; --i >= 0 ; )
	    outc(' ');
    }
}

void
outs(s)
    char *s;
{
    register char *p;

    for (p = s ; *p ; p++)
	outc(*p);
}


void
outc(c)
    char c;
{
    struct outdata *odp;

    odp = osp->data;
    if (odp == NULL || odp->len == sizeof (odp->buf)) {
	odp = (struct outdata *) xmalloc(sizeof (struct outdata));
	odp->len = 0;
	odp->next = osp->data;
	osp->data = odp;
    }
    odp->buf[odp->len++] = c;
}
/* END PACKAGE output */


/* PACKAGE BODY define_queue */

struct queued_define {
    struct queued_define *next;
    char *name;
    char *value;
};

#ifdef TYPE_DEFINE
struct queued_type_define {
    struct queued_type_define *next;
    char *name;
    tree type;
    int unbounded;
};
#endif


struct queued_define *define_queue;
struct queued_define **end_define_queue = &define_queue;

#ifdef TYPE_DEFINE
struct queued_type_define *type_define_queue;
struct queued_type_define **end_type_define_queue = &type_define_queue;
#endif


/*
 * Queue a define for later output.  Both the arguments must point to
 * strings that will stay around.
 */
void
queue_define(name, value)
    char *name;
    char *value;
{
    struct queued_define *p;

    p = (struct queued_define *) xmalloc(sizeof (struct queued_define));
    p->name = name;
    p->value = value;
    *end_define_queue = p;
    end_define_queue = &p->next;
}

#ifdef TYPE_DEFINE
/*
 * Queue a type define for later output.  The first argument must point
 * to a string that will stay around.
 */
void
queue_type_define(name, type, unbounded)
    char *name;
    tree type;
    int unbounded;
{
    struct queued_type_define *p;

    p = (struct queued_type_define *) xmalloc(sizeof (struct queued_type_define));
    p->name = name;
    p->type = type;
    p->unbounded = unbounded;
    *end_type_define_queue = p;
    end_type_define_queue = &p->next;
}
#endif

void
put_other_defines() {
    {
	struct queued_define *p, *next;

	*end_define_queue = NULL;
	for (p = define_queue ; p ; p = next) {
	    next = p->next;
	    fprintf(outfp, "define([%s], [%s])\n", p->name, p->value);
	    free(p);
	}
	end_define_queue = &define_queue;
    }
#ifdef TYPE_DEFINE
    {
	struct queued_type_define *p, *next;

	*end_type_define_queue = NULL;
	for (p = type_define_queue ; p ; p = next) {
	    next = p->next;
	    fprintf(outfp, "define([%s], [", p->name);
	    start_output();
	    push_output();
	    put_type(p->type, NULL, p->unbounded);
	    pop_output();
	    fputs("])\n", outfp);
	    free(p);
	}
	end_type_define_queue = &type_define_queue;
    }
#endif
}
/* END PACKAGE define_queue */
