--------------------------------------------------------------------------
--  Forest library, version 1.4beta					--
--  Copyright (C) 1995 by Kenneth Almquist				--
--									--
--  You may redistribute or modify this software under the terms	--
--  given in the License for the Forest Library, which appears in	--
--  the file named LICENSE.						--
--------------------------------------------------------------------------

with POSIX,
     POSIX_IO, 
     POSIX_Process_Identification;

package POSIX_File_Locking is

    -- The file locking works on regions of a file.  Locking an entire
    -- file (which can be done by setting whole_file to true) is equivalent
    -- to locking an infinitely long region starting at position zero.
    -- It is legal to lock regions which are beyond the end of file.

    type Lock_Kind is (Read_Lock, Write_Lock, Unlock);

    type File_Lock (Whole_File: Boolean := true) is record
        Lock:  Lock_Kind;
        case Whole_File is
	    when true => null;
	    when false =>
		Starting_Point:	 POSIX_IO.Position;
		Start:		 POSIX_IO.IO_Offset;
		Length:		 POSIX.IO_Count;
        end case;
    end record;

    procedure Get_Lock
	    (File:     POSIX_IO.File_Descriptor;
    	     Lock:     File_Lock;
	     Result:   out File_Lock;
             Process:  out POSIX_Process_Identification.Process_ID);
	-- Exceptions:  posix_error.
	-- Sets result to "the first" lock which conflicts with lock,
	-- and process to the process_id of the process holding the
	-- lock.  If there is no conflicting lock, sets process to
	-- null_process_id.

    procedure Set_Lock
	    (File:   POSIX_IO.File_Descriptor;
	     Lock:   File_Lock);
	-- Exceptions:  posix_error.
	-- Set or release a lock, raising posix_error with error_code
	-- Resource_Temporarily_Unavailable if there is a conflicting
	-- lock.
	--
	-- If a process holds one or more locks on a file, and places
	-- a new lock on the file, the new lock overrides the old locks
	-- if the regions covered overlap.  For example, if you say
	--	set_lock(file, (false, write_lock, from_beginning, 3, 5);
	--	set_lock(file, (false, read_lock, from_beginning, 3, 2);
	-- the second set_lock call will place a read lock on the two
	-- byte region starting at position 3, and delete the write lock
	-- on this region.  The next three bytes will continue to be
	-- covered by the write lock from the first set_lock call.

    procedure Wait_To_Set_Lock
	    (File:	      POSIX_IO.File_Descriptor;
	     Lock:	      File_Lock;
             Masked_Signals:  POSIX.More.Signal_Handling := POSIX.More.Restart);
	-- Exceptions:  posix_error.
	-- Set or release a lock, blocking until any conflicting locks
	-- are released.

end POSIX_File_Locking;
