--------------------------------------------------------------------------
--  Forest library, version 1.4beta					--
--  Copyright (C) 1995 by Kenneth Almquist				--
--									--
--  You may redistribute or modify this software under the terms	--
--  given in the License for the Forest Library, which appears in	--
--  the file named LICENSE.						--
--------------------------------------------------------------------------

package POSIX_Private_2.Task_Safety is

    -- Task safety.

    -- Rather than having individual semaphores for various items, we
    -- have a single semaphore that controls all accesses to protected
    -- items.  This is simpler to implement, and should be adequate, at
    -- least on non-multiprocessor machines.  The semaphore supports
    -- three types of locks:  write locks, read locks, and transient
    -- locks.  The concise rule is that if a task holds a write lock,
    -- then no other task may hold a read or write lock, and no task
    -- holding a transient lock may execute.  What this means is that
    -- read and write locks behave in the standard fashion, while
    -- transient locks are read locks that may be released when the
    -- task holding them is not executing.  Transient locks are used
    -- to protect the umask value.
    --
    -- A single task can hold more than one lock.  In particular, a
    -- task holding a write lock can get another write lock.
    --
    -- The expected way to implement this package on a uniprocessor
    -- is to make allocating a read or write lock disable the
    -- scheduler (e.g. by calling pthread_sched_prevent() on the
    -- Provenzano pthreads package), and to make holding a transient
    -- lock do nothing.  This means that a task holding a read or
    -- write lock should not perform task redevous operations or
    -- execute slow system calls, since those might invoke the scheduler.
    -- (Slow system calls are those which have EINTR as one of the
    -- possible errno values.)
    --
    -- A lock handler variable is a controlled type which keeps track
    -- of whether a lock is held.  If the lock is held when the lock
    -- handler is finalized, the lock will be released.  This is useful
    -- for ensuring that a lock is released on exit from a subroutine,
    -- even if the exit is due to an exception.

    type Write_Lock_Holder is limited private;
    type Read_Lock_Holder is limited private;
    type Transient_Lock_Holder is limited private;

    procedure Task_Write_Lock(Holder: in out Write_Lock_Holder);
	-- Exceptions:  none.
	-- Grab a write lock, recording the fact that it is held in the holder.

    procedure Task_Write_Unlock(Holder: in out Write_Lock_Holder);
	-- Exceptions:  none.
	-- Release a write lock, updating the holder.

    procedure Task_Read_Lock(Holder: in out Read_Lock_Holder);
	-- Exceptions:  none.
	-- Grab a read lock, recording the fact that it is held in the holder.

    procedure Task_Read_Unlock(Holder: in out Read_Lock_Holder);
	-- Exceptions:  none.
	-- Release a read lock, updating the holder.

    procedure Task_Transient_Lock(Holder: in out Transient_Lock_Holder);
	-- Exceptions:  none.
	-- Grab a transient lock, recording the fact that it is held.

    procedure Task_Transient_Unlock(Holder: in out Transient_Lock_Holder);
	-- Exceptions:  none.
	-- Release a transient lock, updating the holder.

    pragma inline(Task_Write_Lock);
    pragma inline(Task_Write_Unlock);
    pragma inline(Task_Read_Lock);
    pragma inline(Task_Read_Unlock);
    pragma inline(Task_Transient_Lock);
    pragma inline(Task_Transient_Unlock);


private
    type Write_Lock_Holder is null record;
    type Read_Lock_Holder is null record;
    type Transient_Lock_Holder is null record;

end POSIX_Private_2.Task_Safety;
