--------------------------------------------------------------------------
--  Forest library, version 1.4beta					--
--  Copyright (C) 1995 by Kenneth Almquist				--
--									--
--  You may redistribute or modify this software under the terms	--
--  given in the License for the Forest Library, which appears in	--
--  the file named LICENSE.						--
--------------------------------------------------------------------------

with POSIX,
     POSIX.More,
     POSIX_IO,
     POSIX_Permissions,
     POSIX_Process_Environment,
     POSIX_Process_Identification,
     POSIX_Signals,
     Ada.Finalization;
package POSIX_Process_Primitives is

    -- Process Template

    type Process_Template is limited private;
	-- A process_template contains operations that should be performed
	-- by start_process.  All of the specified operations are performed
	-- in the context of the child process created by start_process.

    procedure Open_Template(Template: in out Process_Template);
	-- Exceptions:  none.
	-- Initialize a process template.  Any operation on a template
	-- besides open_template will raise posix_error (with error_code
	-- set to invalid_argument) if called on an uninitialized template.

    procedure Close_Template(Template: in out Process_Template);
	-- Exceptions:  posix_error
	-- Close a template, freeing all storage associated with it.

    procedure Set_Keep_Effective_IDs(Template: in out Process_Template);
    pragma inline(Set_Keep_Effective_IDs);
	-- Exceptions:  posix_error.
	-- Retain the effective user and group IDs in the child process.
	-- Be sure to call this if you are not writing a set-user-ID or
	-- set-group-ID program.

    procedure Set_Signal_Mask
            (Template:	in out Process_Template;
             Mask:	POSIX_Signals.Signal_Set);
	-- Exceptions:  posix_error.
	-- Specifies the signal mask for the child process.

    procedure Set_Creation_Signal_Masking
            (Template:	     in out Process_Template;
             Masked_Signals: POSIX.More.Signal_Handling := POSIX.RTS_Signals);
	-- Exceptions:  posix_error.
	-- Specifies the signal handling for interruptable operations
	-- performed by start_process.

    procedure Set_File_Action_To_Open
            (Template:	in out Process_Template;
             File:	POSIX_IO.File_Descriptor;
             Name:	POSIX.Pathname;
             Mode:	POSIX_IO.File_Mode := POSIX_IO.Read_Only;
	     Options:	POSIX_IO.Open_Option_Set := POSIX_IO.Empty_Set);
	-- Exceptions:  posix_error.
	-- Causes the child process to open the named file.

    procedure Set_File_Action_To_Close
            (Template: in out Process_Template;
             File:      POSIX_IO.File_Descriptor);
	-- Exceptions:  posix_error.
	-- Causes the child process to close the named file descriptor.

    procedure Set_File_Action_To_Duplicate
            (Template:	 in out Process_Template;
             File:	 POSIX_IO.File_Descriptor;
             From_File:	 POSIX_IO.File_Descriptor);
	-- Exceptions:  posix_error.
	-- Causes the child process to duplicate the named file descriptor.
	-- Operations are performed in the order requested, so to move a
	-- file descriptor you can first duplicate it and then close the
	-- original.


    -- Process Creation

    procedure Start_Process
            (Child:	out POSIX_Process_Identification.Process_ID;
             Pathname:	POSIX.Pathname;
             Template:	Process_Template;
             Env_List:	POSIX_Process_Environment.Environment;
             Arg_List:	POSIX.POSIX_String_List := POSIX.Empty_String_List);
	-- Exceptions:  posix_error.
	-- Start a new process.  A child process is created, the actions
	-- requested by the process template are performed, and the program
	-- specified by pathname is run.  The specified argument list and
	-- environment are passed to the program.  If an error occurs after
	-- the child process is created but before the program has been
	-- loaded successfully, the child process exits with an exit status
	-- of failed_creation_exit.

    procedure Start_Process
            (Child:	out POSIX_Process_Identification.Process_ID;
             Pathname:	POSIX.Pathname;
             Template:	Process_Template;
             Arg_List:	POSIX.POSIX_String_List := POSIX.Empty_String_List);
	-- Exceptions:  posix_error.
	-- As above, but uses the current environment.

    procedure Start_Process_Search
            (Child:	out POSIX_Process_Identification.Process_ID;
             Filename:	POSIX.Filename;
             Template:	Process_Template;
             Env_List:	POSIX_Process_Environment.Environment;
             Arg_List:	POSIX.POSIX_String_List := POSIX.Empty_String_List);
	-- Exceptions:  posix_error.
	-- As above except that if filename does not contain a slash, the
	-- program is searched for using the environment variable PATH.  The
	-- value of PATH is taken from the current environment, not the
	-- environment passed to the program.

    procedure Start_Process_Search
            (Child:	out POSIX_Process_Identification.Process_ID;
             Filename:	POSIX.Filename;
             Template:	Process_Template;
             Arg_List:	POSIX.POSIX_String_List := POSIX.Empty_String_List);
	-- Exceptions:  posix_error.
	-- As above, but uses the current environment.


    -- Process Exit

    type Exit_Status is range 0 .. 255;
	-- Process exit status value.

    Normal_Exit:	       constant Exit_Status := 0;
	-- Exit status indicating success.

    Failed_Creation_Exit:      constant Exit_Status := 41;
	-- See start_process.

    Unhandled_Exception_Exit:  constant Exit_Status := 42;
	-- Indicates that the process terminated due to an unhandled
	-- exception.  (In GNAT 2.03, unhandled exceptions cause the
	-- process to exit with an exit status of 1.  Hopefully this
	-- will be correct in a future release.)

    procedure Exit_Process(Status: Exit_Status := Normal_Exit);
	-- Exceptions:  none.
	-- Terminate the current process, including all tasks.


    -- Termination Status

    type Termination_Status is private;
    type Termination_Cause  is 
            (Exited, Terminated_By_Signal, Stopped_By_Signal);

    function Status_Available(Status: Termination_Status)
        return Boolean;
	-- Exceptions:  none.
	-- Return true if status information is available.  Status
	-- information is available only if wait_for_child_process
	-- returned information on a child process.  If status information
	-- is not available, other operations which extract information
	-- from a termination_status will raise posix_error with error_code
	-- set to invalid_argument.

    function Process_ID_Of(Status: Termination_Status)
        return POSIX_Process_Identification.Process_ID;
    pragma inline(Process_ID_Of);
	-- Exceptions:  posix_error.
	-- Return the process ID of the terminated process.

    function Termination_Cause_Of(Status: Termination_Status)
            return Termination_Cause;
	-- Exceptions:  posix_error.
	-- Return the reason the process terminated.  (If the cause is
	-- stopped_by_signal, the child process is actually stopped
	-- rather than terminated.)

    function Exit_Status_Of(Status: Termination_Status)
        return Exit_Status;
	-- Exceptions:  posix_error.
	-- Returns the exit status.  Termination_cause must be exited.

    function Termination_Signal_Of(Status: Termination_Status)
        return POSIX_Signals.Signal;
	-- Exceptions:  posix_error.
	-- Returns the signal that terminated the process.  Termination_cause
	-- must be terminated_by_signal.

    function Stopping_Signal_Of(Status: Termination_Status)
        return POSIX_Signals.Signal;
	-- Exceptions:  posix_error.
	-- Returns the signal that stopped the process.  Termination_cause
	-- must be stopped_by_signal.

    -- Wait for Process Termination

    procedure Wait_For_Child_Process
            (Status:	      out Termination_Status;
             Child:	      POSIX_Process_Identification.Process_ID;
             Block:	      Boolean:= true;
             Trace_Stopped:   Boolean := true;
             Masked_Signals:  POSIX.More.Signal_Handling := POSIX.More.Restart);
	-- Exceptions:  posix_error.
	-- Place the termination status of the specified child process in the
	-- variable status.
	-- If block is true, then if no termination status is immediately
	-- available, the routine will block until it becomes available.
	-- Otherwise, the routine will return immediately if no status is
	-- available, and exit_status_available(status) will return false.
	-- If trace_stopped is true, then this routine may return a status
	-- with the termination cause stopped_by_signal.  Otherwise, it will
	-- only return status values with the termination cause set to exited
	-- or terminated_by_signal.

    procedure Wait_For_Child_Process
            (Status:	      out Termination_Status;
             Group:	      POSIX_Process_Identification.Process_Group_ID;
             Block:	      Boolean := true;
             Trace_Stopped:   Boolean := true;
             Masked_Signals:  POSIX.More.Signal_Handling := POSIX.More.Restart);
	-- Exceptions:  posix_error.
	-- As above, but get status from any child process in the specified
	-- process group.

    procedure Wait_For_Child_Process
            (Status:	      out Termination_Status;
             Block:	      Boolean := true;
             Trace_Stopped:   Boolean := true;
             Masked_Signals:  POSIX.More.Signal_Handling := POSIX.More.Restart);
	-- Exceptions:  posix_error.
	-- As above, but get status from any child process.


private
    use POSIX;
    use POSIX_Process_Identification;
    type POSIX_String_Ptr is access POSIX_String;
    type File_Action is (Open, Close, Duplicate);
    type File_Request;
    type File_Request_Ptr is access File_Request;
    type File_Request (Action: File_Action) is record
	Next:  File_Request_Ptr;
	File:  POSIX_IO.File_Descriptor;
	case Action is
	    when Open =>
		Name:	    POSIX_String_Ptr;
		Mode:	    POSIX_IO.File_Mode;
		Options:    POSIX_IO.Open_Option_Set;
	    when Close =>
		null;
	    when Duplicate =>
		From_File:  POSIX_IO.File_Descriptor;
	end case;
    end record;
    type Process_Template is new Ada.Finalization.Limited_Controlled with record
	Is_Open:	     Boolean := false;
	Keep_Effective_IDs:  Boolean;
	Signal_Mask:	     POSIX_Signals.Signal_Set;
	Signal_Handling:     POSIX.More.Signal_Handling;
	File_Request_List:   File_Request_Ptr;
    end record;

    procedure Finalize(Template: in out Process_Template);

    type Exit_Stat is mod 2 ** Integer'size;
    type Termination_Status is record
	Pid:	      Process_ID := Null_Process_ID;
	Exit_Status:  Exit_Stat;
    end record;

end POSIX_Process_Primitives;
