--------------------------------------------------------------------------
--  Copyright (C) 1995 by Kenneth Almquist				--
--									--
--  This program is free software; you can redistribute it and/or	--
--  modify it under the terms of the GNU General Public License as	--
--  published by the Free Software Foundation; either version 2 of	--
--  the License, or (at your option) any later version.			--
--------------------------------------------------------------------------

-- Test posix_file_io package.

with test_pkg; use test_pkg;
with text_io; use text_io;
with posix; use posix;
with posix.more; use posix.more;
with posix_permissions; use posix_permissions;
with posix_files; use posix_files;
with posix_files.more; use posix_files.more;
with posix_file_status; use posix_file_status;
with posix_file_status.more; use posix_file_status.more;
with posix_io; use posix_io;
--with posix_process_identification; use posix_process_identification;
--with posix_process_identification.more; use posix_process_identification.more;
with posix_process_environment; use posix_process_environment;
--with posix_process_environment.more; use posix_process_environment.more;
--with posix_process_primitives; use posix_process_primitives;
--with posix_unsafe_process_primitives; use posix_unsafe_process_primitives;
--with posix_file_locking;
--with forest_file_locking;
--with posix_signals; use posix_signals;
--with posix_calendar; use posix_calendar;
--with posix_user_database; use posix_user_database;
--with posix_group_database; use posix_group_database;
with posix_configurable_file_limits; use posix_configurable_file_limits;
--with ada.calendar; use ada.calendar;
with ada.io_exceptions;
with ada.exceptions;
with ada.unchecked_deallocation;
procedure file_io is
    test_fd: file_descriptor;	-- file descriptor for test_file.
    out_fd: file_descriptor;
    in_fd: file_descriptor;
    fd: file_descriptor;
    st1, st2: status;
    last: io_count;
    offset: io_offset;
    options: open_option_set;
    io_mode: posix_io.file_mode;
    nl_char: constant posix_character
		:= posix_character'val(10);	-- newline character
    end_error: exception renames ada.io_exceptions.end_error;
    read_write_perms: constant permission_set
		:= (owner_read | owner_write => true,
		    group_read | group_write => true,
		    others_read | others_write => true,
		    others => false);
begin
    test("O10");
    require(posix_io.standard_input = 0);
    require(posix_io.standard_output = 1);
    require(posix_io.standard_error = 2);

    -- Test open, close, get_file_status, and open_or_create.
    test("O11");
    test_fd := open("test_file", read_write);
    test("O12");
    st1 := get_file_status(test_fd);
    st2 := get_file_status("test_file");
    require(st1 = st2);
    test("O12");
    close(test_fd);
    test("O13");
    begin
	st1 := get_file_status(test_fd);
	except;
    exception
	when x: posix_error =>
	    check_error(bad_file_descriptor, x);
    end;
    test("O14");
    begin
	close(test_fd);
	except;
    exception
	when x: posix_error =>
	    check_error(bad_file_descriptor, x);
    end;
    test("O15");
    begin
	fd := open("nonexistent_file", read_only);
	except;
    exception
	when x: posix_error =>
	    check_error(no_such_file_or_directory, x);
    end;
    test("O20");
    out_fd := open_or_create("outfile", write_only, read_write_perms);
    test("O21");
    st1 := get_file_status("outfile");
    require(is_regular_file(st1));
    require(size_of(st1) = 0);
    require(permission_set_of(st1) = permission_set'(owner_read | owner_write | group_read | others_read => true, others => false));
    st2 := get_file_status(out_fd);
    require(st2 = st1, "Status structures differ");
    test("O22");
    begin
	out_fd := open_or_create("outfile", write_only, read_write_perms,
			exclusive);
	except;
    exception
	when x: posix_error =>
	    check_error(file_exists, x);
    end;
    test("O23");
    out_fd := open_or_create("outfile", write_only, read_write_perms);

    -- Test read and write.
    test("O30");
    test_fd := open("test_file", read_only);
    test("O31");
    declare
	buf: posix_string(1 .. 4);
    begin
	read(test_fd, buf, last);
	require(buf(1 .. integer(last)) = "hell");
    end;
    test("O32");
    declare
	buf: posix_string(101 .. 104);
    begin
	read(test_fd, buf, last);
	require(buf(101 .. integer(last)) = posix_string'('o', nl_char));
    end;
    test("O33");
    declare
	buf: posix_string(1 .. 4);
    begin
	read(test_fd, buf, last);
	except;
    exception
	when end_error => null;
    end;
    test("O34");
    declare
	buf: posix_string(1 .. 4);
    begin
	read(out_fd, buf, last);	-- Try to read a write_only file
	except;
    exception
	when x: posix_error =>
	    check_error(bad_file_descriptor, x);
    end;
    test("O35");
    write(out_fd, "0123456789", last);
    require(last = 10);
    st1 := get_file_status(out_fd);
    require(size_of(st1) = 10);
    test("O36");
    declare
	buf: posix_string := "xxx";
    begin
	write(test_fd, buf, last);	-- Try to write a read_only file
	except;
    exception
	when x: posix_error =>
	    check_error(bad_file_descriptor, x);
    end;

    -- Test seek, file_size, and file_position.
    test("O40");
    require(file_size(test_fd) = 6);
    require(file_size(out_fd) = 10);
    test("O41");
    require(file_position(test_fd) = 6);
    test("O42");
    seek(test_fd, 1, offset, from_beginning);
    require(offset = 1);
    declare
	buf: posix_string(1 .. 1);
    begin
	read(test_fd, buf, last);
	require(last = 1);
	require(buf(1) = 'e');
    end;
    require(file_position(test_fd) = 2);
    test("O43");
    seek(test_fd, 1, offset, from_current_position);
    require(offset = 3);
    require(file_position(test_fd) = 3);
    test("O44");
    seek(test_fd, -1, offset, from_end_of_file);
    require(offset = 5);
    require(file_position(test_fd) = 5);
    test("O45");
    close(test_fd);
    begin
	seek(test_fd, 1, offset, from_beginning);
	except;
    exception
	when x: posix_error =>
	    check_error(bad_file_descriptor, x);
    end;

    -- Test generic_read and generic_write.
    -- At this point, test_fd is closed and out_fd refers to "outfile"
    test("O46");
    declare
	type string_4 is new posix_string(1 .. 4);
	procedure read_4 is new generic_read(string_4);
	buf: string_4;
	procedure read_bool is new generic_read(boolean);
	procedure write_bool is new generic_write(boolean);
	fd: file_descriptor;
	b: boolean;
    begin
	test_fd := open("test_file", read_only);
	test("O850");
	read_4(test_fd, buf);
	require(buf = "hell");
	test("O850");
	begin
	    read_4(test_fd, buf);
	    except;
	exception
	    when end_error => null;
	end;
	test("O852");
	fd := open_or_create("booleans", write_only,
			(owner_read | owner_write => true, others => false),
			exclusive);
	test("O853");
	write_bool(fd, true);
	write_bool(fd, false);
	test("O854");
	close(fd);
	fd := open("booleans", read_only);
	test("O855");
	read_bool(fd, b);
	require(b = true, "true expected");
	read_bool(fd, b);
	require(b = false, "true expected");
	test("O856");
	begin
	    read_bool(fd, b);
	    except;
	exception
	    when end_error => null;
	end;
	test("O857");
	close(fd);
	unlink("booleans");
    exception
	when occurrence: posix_error =>
	    fail("Unexpected POSIX_Error: "
		    & ada.exceptions.exception_message(occurrence));
    end;
    close(test_fd);
    test("O860");
    declare
	-- Test whether generic_read and generic_write can handle the
	-- following type:
	type b510 is array (natural range 0 .. 509) of boolean;
	pragma pack(b510);
	procedure read_b510 is new generic_read(b510);
	procedure write_b510 is new generic_write(b510);
	type use_b510 is record
	    a: b510;
	    b: boolean;
	end record;
	-- The following clause is rejected by GNAT 3.01.
	-- for use_b510 use record
	--     a at 0 range 0 .. 509;
	--     b at 63 range 6 .. 6;
	-- end record;
	pragma pack(use_b510);
	r: use_b510 := ((others => false), false);
	fd: file_descriptor;
	file: constant posix_string (3 .. 10) := "booleans";
    begin
	fd := open_or_create(file, write_only,
			(owner_read | owner_write => true, others => false),
			exclusive, all_signals);
	test("O861");
	r.a(509) := false;
	r.b := false;
	write_b510(fd, r.a, all_signals);
	close(fd);
	test("O862");
	fd := open(file, read_only, empty_set, all_signals);
	test("O863");
	r.a(509) := true;
	r.b := true;
	read_b510(fd, r.a, all_signals);
	require(r.a(509) = false, "r.a(510) not read");
	require(r.b = true, "r.b overwritten");
	test("O864");
	close(fd, all_signals);
	unlink(file);
    end;

    -- Test file_control, close_on_exec, and is_open.
    test("O50");
    get_file_control(out_fd, io_mode, options);
    require(io_mode = write_only);
    require(options = empty_set);
    test("O51");
    set_file_control(out_fd, posix_io.append);
    seek(out_fd, 0, offset, from_beginning);
    declare
	buf: posix_string := "x";
    begin
	write(out_fd, buf, last);
	require(last = 1);
    end;
    require(file_position(out_fd) = 11);
    test("O52");
    get_file_control(out_fd, io_mode, options);
    require(io_mode = write_only);
    require(options = posix_io.append);
    test("O54");
    require(get_close_on_exec(out_fd) = false);
    test("O55");
    set_close_on_exec(out_fd, true);
    test("O56");
    require(get_close_on_exec(out_fd) = true);
    test("O57");
    require(is_open(out_fd));
    test("O60");
    close(out_fd);	-- cause subsequent operations to fail
    begin
	get_file_control(out_fd, io_mode, options);
	except;
    exception
	when x: posix_error =>
	    check_error(bad_file_descriptor, x);
    end;
    test("O61");
    begin
	set_file_control(out_fd, empty_set);
	except;
    exception
	when x: posix_error =>
	    check_error(bad_file_descriptor, x);
    end;
    test("O62");
    declare
	flag: boolean;
    begin
	flag := get_close_on_exec(out_fd);
	except;
    exception
	when x: posix_error =>
	    check_error(bad_file_descriptor, x);
    end;
    test("O63");
    begin
	set_close_on_exec(out_fd, false);
	except;
    exception
	when x: posix_error =>
	    check_error(bad_file_descriptor, x);
    end;
    test("O64");
    require(is_open(out_fd) = false);
    require(is_open(test_fd) = false);
    test("O65");
    out_fd := open("outfile", read_write);
    get_file_control(out_fd, io_mode, options);
    require(io_mode = read_write);

    -- Test duplication operations and pipes.
    test("O70");
    test_fd := open("test_file", read_only);
    fd := duplicate(test_fd);
    require(fd /= test_fd);
    st1 := get_file_status(fd);
    st2 := get_file_status(test_fd);
    require(file_id_of(st1) = file_id_of(st2) and
      device_id_of(st1) = device_id_of(st2));
    test("O71");
    if duplicate_and_close(out_fd, fd) /= fd then
	fail("wrong duplicate");
    end if;
    st1 := get_file_status(fd);
    st2 := get_file_status(out_fd);
    require(file_id_of(st1) = file_id_of(st2) and
      device_id_of(st1) = device_id_of(st2));
    close(fd);
    close(out_fd);
    test("O72");
    create_pipe(in_fd, out_fd);
    test("O73");
    declare
	outbuf: posix_string := "AB";
	inbuf: posix_string(1 .. 2);
    begin
	write(out_fd, outbuf, last);
	require(last = 2);
	read(in_fd, inbuf, last);
	require(last = 2);
	require(inbuf = outbuf);
    end;

    -- Test nonblocking I/O and Transfer_Count_Of.
    -- in_fd and out_fd refer to a pipe
    test("O890");
    set_file_control(in_fd, non_blocking);
    declare
	buf: posix_string(1 .. 2);
    begin
	read(in_fd, buf, last);
	require(last = 0);
    end;
    test("O891");
    declare
	type string_4 is new posix_string (1 .. 4);
	procedure read_4 is new generic_read(string_4);
	outbuf: posix_string (1 .. 3) := "ABC";
	buf: string_4;
    begin
	write(out_fd, outbuf, last);
	require(last = 3);
	test("O892");
	begin
	    read_4(in_fd, buf);
	exception
	    when x : posix_error =>
		require(transfer_count_of(x) = 3, "wrong transfer count");
		check_error(resource_temporarily_unavailable, x);
	end;
    end;
    test("O893");
    set_file_control(out_fd, non_blocking);
    declare
	pipe_size: constant io_count := pipe_length_limit(out_fd);
	type big is new posix_string (1 .. integer(pipe_size) + 1);
	type big_ptr is access big;
	procedure big_write is new posix_io.generic_write(big);
	bigbuf: big_ptr;
	procedure free is new ada.unchecked_deallocation(big, big_ptr);
    begin
	bigbuf := new big'(others => 'X');
	begin
	    big_write(out_fd, bigbuf.all);
	exception
	    when x : posix_error =>
		require(transfer_count_of(x) = pipe_size,
			"wrong transfer count");
		check_error(resource_temporarily_unavailable, x);
	end;
	free(bigbuf);
    end;

    set_file_control(in_fd, empty_set);

    -- Test failed seek.
    -- in_fd refers to a pipe.
    test("O75");
    begin
	seek(in_fd, 1, offset, from_beginning);
	except;
    exception
	when x: posix_error =>
	    check_error(invalid_seek, x);
    end;
    close(in_fd);
    close(out_fd);

    -- Test tty functions.
    test("O80");
    require(is_a_terminal(posix_io.standard_error));
    test("O81");
    require(not is_a_terminal(test_fd));
    test("O82");
    declare
	tty_name: posix_string := get_terminal_name(posix_io.standard_error);
    begin
	put("tty name = "); put(to_string(tty_name)); nl;
	require(tty_name = environment_value_of("TTY_NAME"));
    end;
    close(test_fd);
exception
    when occurrence: posix_error =>
	fail("Unexpected POSIX_Error: "
		& ada.exceptions.exception_message(occurrence));
end file_io;
