--------------------------------------------------------------------------
--  Copyright (C) 1995 by Kenneth Almquist				--
--									--
--  This program is free software; you can redistribute it and/or	--
--  modify it under the terms of the GNU General Public License as	--
--  published by the Free Software Foundation; either version 2 of	--
--  the License, or (at your option) any later version.			--
--------------------------------------------------------------------------

with test_pkg; use test_pkg;
with text_io; use text_io;
with ada.integer_text_io; use ada.integer_text_io;
with posix; use posix;
with posix_process_identification; use posix_process_identification;
with posix_signals; use posix_signals;
procedure signals is
    sig: signal;
    set: signal_set;
    original_mask: signal_set;
    was, prev: signal_set;

    -- Verify that a signal set is empty.
    procedure check_empty(sigset: signal_set) is
    begin
	for s in signal'range loop
	    if is_member(sigset, s) then
		fail("Signal set contains" & signal'image(s));
		return;
	    end if;
	end loop;
    end check_empty;

    -- Print the contents of a signal set.
    procedure put_signal_set(description: string; set: signal_set) is
    begin
	put_line(description);
	for s in signal'range loop
	    declare
		name: constant string := image(s);
	    begin
		put("    ");
		put(name);
		put(':');
		for i in name'length + 1 .. 32 loop
		    put(' ');
		end loop;
	    end;
	    put(boolean'image(is_member(set, s)));
	    new_line;
	end loop;
	flush;
    end put_signal_set;
begin
    -- Test Image and Value for signals.
    test("SG10");
    declare
	s: constant string := image(signal_alarm);
    begin
	require(s = "SIGNAL_ALARM", "wrong value");
	require(s'first = 1, "wrong s'first");
    end;
    test("SG11");
    declare
	s: constant string := image(signal_null);
    begin
	require(s = "SIGNAL_NULL", "wrong value");
	require(s'first = 1, "wrong s'first");
    end;
    test("SG12");
    sig := value("SIGNAL_NULL");
    require(sig = signal_null);
    test("SG13");
    sig := value("Signal_Hangup");
    require(sig = signal_hangup);
    test("SG14");
    sig := value("  signal_terminate  ");
    require(sig = signal_terminate);
    test("SG15");
    sig := value("  signull  ");
    require(sig = signal_null);
    test("SG16");
    sig := value("SIGUSR1");
    require(sig = signal_user_1);
    test("SG17");
    begin
	sig := value("No_Such_Signal");
	except;
    exception
	when constraint_error => null;
    end;
    test("SG18");
    begin
	sig := value("");
	except;
    exception
	when constraint_error => null;
    end;
    test("SG19");
    for s in signal'range loop
	if value(image(s)) /= s then
	    put("image("); put(integer(s), width => 1); put(") = ");
	    put(image(s)); new_line;
	    fail("value(image(s)) /= s");
	end if;
    end loop;

    -- Test signal sets.
    test("SG20");
    check_empty(set);
    test("SG21");
    add_signal(set, signal_null);
    require(is_member(set, signal_null));
    test("SG22");
    add_signal(set, signal_user_1);
    require(is_member(set, signal_user_1));
    test("SG23");
    delete_signal(set, signal_null);
    require(not is_member(set, signal_null));
    test("SG24");
    delete_signal(set, signal_user_1);
    require(not is_member(set, signal_user_1));
    test("SG25");
    add_all_signals(set);
    for s in signal'range loop
	require(is_member(set, s));
    end loop;
    test("SG26");
    delete_all_signals(set);
    check_empty(set);

    -- Test signal masking
    test("SG30");
    delete_all_signals(set);
    add_signal(set, signal_user_1);
    set_blocked_signals(set, original_mask);
    test("SG31");
    set := blocked_signals;
    was := set;
    require(is_member(set, signal_user_1));
    delete_signal(set, signal_user_1);
    check_empty(set);
    test("SG32");
    add_signal(set, signal_user_2);
    block_signals(set, prev);
    require(prev = was, "Previous signal mask wrong");
    add_signal(set, signal_user_1);
    was := blocked_signals;
    require(was = set, "Signal not added correctly");
    test("SG33");
    delete_signal(set, signal_user_1);  -- Set now contains signal_user_2 only
    unblock_signals(set, prev);
    require(prev = was, "Previous signal mask wrong");
    was := blocked_signals;
    require(not is_member(was, signal_user_2), "SIGUSR2 not unblocked");
    require(is_member(was, signal_user_1), "SIGUSR1 not blocked");

    -- Test Ignore, send_signal, and pending_signals.
    -- Set_Signal is also tested in the process primitives tests.
    test("SG40");
    require(not is_ignored(signal_user_1));
    test("SG41");
    set := pending_signals;
    check_empty(set);
    test("SG42");
    send_signal(get_process_id, signal_user_1);
    test("SG43");
    set := pending_signals;
    require(is_member(set, signal_user_1));
    test("SG44");
    ignore_signal(signal_user_1);
    test("SG45");
    -- The pending signal should have been cleared when the signal
    -- was ignored.
    set := pending_signals;
    check_empty(set);
    test("SG46");
    require(is_ignored(signal_user_1));
    test("SG47");
    unignore_signal(signal_user_1);
    require(not is_ignored(signal_user_1));
    test("SG50");
    set_blocked_signals(original_mask, prev);
end signals;
