--------------------------------------------------------------------------
--  Copyright (C) 1995 by Kenneth Almquist				--
--									--
--  This program is free software; you can redistribute it and/or	--
--  modify it under the terms of the GNU General Public License as	--
--  published by the Free Software Foundation; either version 2 of	--
--  the License, or (at your option) any later version.			--
--------------------------------------------------------------------------

with posix.more,
     posix_configurable_system_limits,
     posix_configurable_file_limits,
     posix_io,
     ada.text_io,
     ada.integer_text_io;
procedure sysinfo is

    -- Use clauses.

    use posix, posix.more;
    use posix_configurable_system_limits;
    use posix_configurable_file_limits;
    use posix_io;
    use ada.text_io;
    use ada.integer_text_io;

    -- Constants.

    root: constant posix_string := "/";

    -- Variables.

    fd: file_descriptor;
    fd_valid: boolean;
    -- Support procedures.

    procedure put_version(item: posix_version) is
	result: constant string := posix_version'image(item);
	first: natural;
    begin
	first := result'first;
	if result(first) = ' ' then
	    first := first + 1;
	end if;
	put(result(first .. result'last - 2));
	put('_');
	put(result(result'last - 1 .. result'last));
    end put_version;
    
    procedure put_sys_limit(text: string; value: integer) is
    begin
	put(text); put(value); new_line;
    end put_sys_limit;

    procedure put_support(text: string; value: boolean) is
    begin
	put(text);
	if value then
	    put(" is supported.");
	else
	    put(" is not supported.");
	end if;
	new_line;
    end put_support;

    -- Start of dummy limit used for testing --
    subtype dummy_maxima is natural range 1 .. 10000;
    function dummy_is_limited(fd: file_descriptor) return boolean is
    begin
	return false;
    end;
    function dummy_limit(fd: file_descriptor) return dummy_maxima is
    begin
	return 10000;
    end;
    function dummy_is_limited(name: posix_string) return boolean is
    begin
	return dummy_is_limited(0);
    end;
    function dummy_limit(name: posix_string) return dummy_maxima is
    begin
	return 100;
	--return dummy_limit(0);
    end;
    -- End of dummy limit used for testing --

    generic
	text: in string;
	type maxima is new integer;
	with function name_is_limited(name: posix_string) return boolean;
	with function name_limit(name: posix_string) return maxima;
	with function fd_is_limited(fd: file_descriptor) return boolean;
	with function fd_limit(fd: file_descriptor) return maxima;
    procedure put_file_limit;
    procedure put_file_limit is
	is_limited: boolean;
	limit: maxima;
    begin	
	put(text);
	begin
	    is_limited := fd_is_limited(fd);
	exception
	    when x : posix_error =>
		put("Cannot determine, ");
		put(error_string(error_code_of(x)));
		new_line;
		return;
	end;
	limit := fd_limit(fd);
	if is_limited then
	    put(integer(limit));
	else
	    put("  Unlimited");
	end if;
	new_line;
	if name_is_limited(root) /= is_limited
		or name_limit(root) /= limit then
	    put("ERROR: Inconsistent limits"); new_line;
	elsif (not is_limited) and limit /= maxima'last then
	    put("ERROR: Limit /= maxima'last"); new_line;
	end if;
    end;

    procedure put_link_limit is new put_file_limit(
	"Max links to a file:         ", link_limit_maxima,
	link_is_limited, link_limit,
	link_is_limited, link_limit);
    procedure put_line_limit is new put_file_limit(
	"Max length of tty input line:", input_line_limit_maxima,
	input_line_is_limited, input_line_limit,
	input_line_is_limited, input_line_limit);
    procedure put_queue_limit is new put_file_limit(
	"Size of tty input queue:     ", input_queue_limit_maxima,
	input_queue_is_limited, input_queue_limit,
	input_queue_is_limited, input_queue_limit);
    procedure put_filename_limit is new put_file_limit(
	"Max length of file name:     ", filename_limit_maxima,
	filename_is_limited, filename_limit,
	filename_is_limited, filename_limit);
    procedure put_pathname_limit is new put_file_limit(
	"Max length of pathname:      ", pathname_limit_maxima,
	pathname_is_limited, pathname_limit,
	pathname_is_limited, pathname_limit);
    procedure put_pipe_limit is new put_file_limit(
	"Max data buffered in a pipe: ", pipe_limit_maxima,
	pipe_length_is_limited, pipe_length_limit,
	pipe_length_is_limited, pipe_length_limit);
    procedure put_dummy_limit is new put_file_limit(
	"Test value:                  ", dummy_maxima,
	dummy_is_limited, dummy_limit,
	dummy_is_limited, dummy_limit);

    procedure put_feature(have_it: string; do_not_have_it: string;
			  fd_value: boolean; name_value: boolean) is
    begin
	if fd_value then
	    put(have_it);
	else
	    put(do_not_have_it);
	end if;
	new_line;
	if name_value /= fd_value then
	    put("ERROR: inconsistent support values"); new_line;
	end if;
    end put_feature;

begin
    put("You are running "); put(to_string(system_name));
    put(", release "); put(to_string(release));
    put(", version "); put(to_string(version));
    new_line;
    put("Hardware:  "); put(to_string(machine));
    new_line;
    put("The system is named "); put(to_string(node_name));
    new_line;
    put("The system POSIX version is ");
    put_version(posix_configurable_system_limits.system_posix_version);
    new_line;
    put("The version of POSIX.5 that this binding is based on is ");
    put_version(system_posix_ada_version);
    new_line(2);
    put_sys_limit("Argument/environment limit:  ", argument_list_maximum);
    put_sys_limit("Max processes per user:      ", child_processes_maximum);
    put_sys_limit("Max suplementary groups:     ", groups_maximum);
    put_sys_limit("Max C language streams:      ", stream_maximum);
    put_sys_limit("Max length of time zone:     ", time_zone_string_maximum);
    put_support("Job control", job_control_supported);
    put_support("The saved IDs kludge", saved_ids_supported);

    fd_valid := true;
    begin
	fd := open(root, read_only, empty_set, restart);
    exception
	when posix_error =>
	    put("Could not open """ & to_string(root) & """"); new_line;
	    fd_valid := false;
    end;
    if fd_valid then
	new_line;
	put_link_limit;
	put_line_limit;
	put_queue_limit;
	put_filename_limit;
	put_pathname_limit;
	put_pipe_limit;
	-- put_dummy_limit;
	put_feature("Change_Owner is restricted.",
		    "Change_Owner is unrestricted.",
		    change_owner_is_restricted(fd),
		    change_owner_is_restricted(root));
	put_feature("File names longer than the limit are truncated.",
		    "File names longer than the limit raise POSIX_Error.",
		    filename_is_truncated(fd),
		    filename_is_truncated(root));
    end if;
end sysinfo;
